/*	The prupose of this exercise is, to use Taylor Series
 * Expansion, and Common Sense, to define basic trig functions,
 * on a CPU that has floating-point capabilities, and in C++,
 * but, as though trig functions were not available as-is.
 *
 * March 6, 2022
 * Dirk Mittler
 */

#include "functions.h"

#define DIGITS2 std::numeric_limits<double>::digits

double TSin1(double x);
double TAsin1(double x);

const double FP_MIN = 1.0 / (1UL << (DIGITS2 / 5));
const double double_nu = std::numeric_limits<double>::epsilon();
const double FP_PREC = double_nu;

/*	The first fact which must be respected is, that Taylor Series
 * will not be accurate or feasible, to compute Sine functions
 * outside -M_PI and +M_PI, and that it will not be feasible to
 * compute Arc-Sine functions outside -sqrt(0.5) and +sqrt(0.5).
 */

const double Root_Half = sqrt(0.5);

double TSin(const double &x)
{
    int temp = floor((x + M_PI) / (M_PI * 2));
    return TSin1(x - ((temp << 1) * M_PI) - ((temp) * double_nu));
}

/*
double TSin(const double &x) {
	double z = x;
	while (z > M_PI) z -= M_PI;
	while (z < -M_PI) z += M_PI;
	return TSin1(z);
}
*/

double TCos(const double &x)
{
    double temp = M_PI_2 - x;
    return TSin(temp);
}

double TTan(const double &x)
{
    return TSin(x) / TCos(x);
}

double TAsin(const double &x)
{

    if (abs(x) > 1.0) throw TFPE();

    if (x > Root_Half) {
        return M_PI_2 - TAsin1(sqrt(1 - (x * x)));
    } else if (x >= 0) {
        return TAsin1(x);
    } else if (x >= -Root_Half) {
        return -TAsin1(-x);
    }

    return -M_PI_2 + TAsin1(sqrt(1 - (x * x)));

}

double TAcos(const double &x)
{
    return M_PI_2 - TAsin(x);
}

double TAtan(const double &x)
{
    double temp = x / sqrt(1 + (x * x));
    return TAsin(temp);
}

/*	Now, let's get to work... */

double TSin1(double x)
{
    double additive = 4.0;
    double exponent = x;
    double factorial = 1.0;
    double accum = 0.0;
    int iter = 1;
    bool simplify = false;
    int in_sign = (x >= 0) ? 1 : -1;
    double temp_sine = 1.0;

    if (abs(x) < FP_MIN * 2) simplify = true;

    while (abs(additive) > abs(accum) * FP_PREC) {

        accum += factorial * exponent;

        factorial /= iter + 1;
        factorial /= iter + 2;
        exponent *= x * x;

        additive = factorial * exponent;
        accum -= additive;

        if (iter > 3 && (simplify || abs(accum) < FP_MIN)) break;
        if (iter > 5 && abs(1 - abs(accum)) < FP_MIN &&
                temp_sine > Root_Half) {
            temp_sine = abs(TSin1(x - (M_PI_2 * in_sign)));
            if (temp_sine < FP_MIN) break;
        }

        exponent *= x * x;

        factorial /= iter + 3;
        factorial /= iter + 4;
        iter += 4;
    }

    if (!simplify && abs(accum) < FP_MIN) {
        accum = -TSin1(x - (M_PI * in_sign));
    } else if (temp_sine < FP_MIN) {
        accum = sqrt(1 - (temp_sine * temp_sine)) * in_sign;
    }

    if (abs(accum) == 0.0) accum = 0.0;

    return accum;
}

double TAsin1(double x)
{
    double additive = 1.0;
    double exponent = x;
    double factorial = 1.0;
    double accum = 0.0;
    int iter = 1;
    bool simplify = false;

    if (abs(x) < FP_MIN) simplify = true;

    while (abs(additive) > abs(accum) * FP_PREC) {

        additive = factorial * exponent / iter;
        accum += additive;

        if (iter > 3 && simplify) break;

        exponent *= x * x;

        factorial *= double(iter) / (iter + 1);
        iter += 2;
    }

    return accum;
}
